/*
  Copyright (C) 2000-2008

  Code contributed by Greg Collecutt, Joseph Hope and Paul Cochrane

  This file is part of xmds.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
  $Id: xmds_integrate_ark45ex.cc 1885 2008-03-18 15:24:56Z paultcochrane $
*/

/*! @file xmds_integrate_ark45ex.cc
  @brief Integrate element parsing classes and methods; fourth order Runge-Kutta in the explicit picture

  More detailed explanation...
*/

#include <xmds_common.h>
#include <xmds_integrate_ark45ex.h>
#include <xmds_vector.h>
#include <xmds_simulation.h>

// **************************************************************************
// **************************************************************************
//                              xmdsIntegrateARK45EX public
// **************************************************************************
// **************************************************************************

extern bool debugFlag;

long nxmdsIntegrateARK45EXs=0;  //!< The number of integrate ARK45EX objects

// **************************************************************************
xmdsIntegrateARK45EX::xmdsIntegrateARK45EX(
                                           const xmdsSimulation *const yourSimulation,
                                           const bool& yourVerboseMode) :
  xmdsIntegrate(yourSimulation, yourVerboseMode, false, false),
  xmdsIntegrateEX(yourSimulation, yourVerboseMode),
  xmdsIntegrateARK45(yourSimulation, yourVerboseMode) {
  if (debugFlag) {
    nxmdsIntegrateARK45EXs++;
    printf("xmdsIntegrateARK45EX::xmdsIntegrateARK45EX\n");
    printf("nxmdsIntegrateARK45EXs=%li\n", nxmdsIntegrateARK45EXs);
  }
}

// **************************************************************************
xmdsIntegrateARK45EX::~xmdsIntegrateARK45EX() {
  if (debugFlag) {
    nxmdsIntegrateARK45EXs--;
    printf("xmdsIntegrateARK45EX::~xmdsIntegrateARK45EX\n");
    printf("nxmdsIntegrateARK45EXs=%li\n", nxmdsIntegrateARK45EXs);
  }
}

// **************************************************************************
void xmdsIntegrateARK45EX::processElement(
                                          const Element *const yourElement) {
  if (debugFlag) {
    printf("xmdsIntegrateARK45EX::processElement\n");
  }

  if (verbose()) {
    printf("Processing integrate ARK45EX element ...\n");
  }

  xmdsIntegrate::processElement(yourElement);
  xmdsIntegrateARK45::processElement(yourElement);
  xmdsIntegrateEX::processElement(yourElement);
}

// **************************************************************************
// **************************************************************************
//                              xmdsIntegrateARK45EX private
// **************************************************************************
// **************************************************************************

// **************************************************************************
void xmdsIntegrateARK45EX::writePrototypes(
                                           FILE *const outfile) const {
  if (debugFlag) {
    printf("xmdsIntegrateARK45EX::writePrototypes\n");
  }

  const xmdsVector* mainVector;

  if (!simulation()->field()->getVector("main", mainVector)) {
    throw xmdsException("Internal error in xmdsIntegrateARKIP::writeResetRoutine: cannot find 'main' vector");
  }

  const char* typeName="";
  if (mainVector->vectorType()==COMPLEX) {
    typeName="complex";
  }
  else if (mainVector->vectorType()==DOUBLE) {
    typeName="double";
  }

  fprintf(outfile, "// ********************************************************\n");
  fprintf(outfile, "// segment %li (ARK45EX) prototypes\n", segmentNumber);
  fprintf(outfile, "\n");
  xmdsIntegrate::writePrototypes(outfile);
  xmdsIntegrateEX::writePrototypes(outfile);

  fprintf(outfile, "\n");
  fprintf(outfile, "// integrate (ARK45EX) prototypes\n");
  fprintf(outfile, "\n");
  fprintf(outfile, "void _segment%li_reset(%s* _reset_to, double _step);\n", segmentNumber, typeName);
  xmdsIntegrateARK45::writePrototypes(outfile);
}

// **************************************************************************
void xmdsIntegrateARK45EX::writeRoutines(
                                         FILE *const outfile) const {
  if (debugFlag) {
    printf("xmdsIntegrateARK45EX::writeRoutines\n");
  }

  fprintf(outfile, "// ********************************************************\n");
  fprintf(outfile, "// segment %li (ARK45EX) routines\n", segmentNumber);
  fprintf(outfile, "\n");

  xmdsIntegrate::writeRoutines(outfile);
  writeResetRoutine(outfile);
  xmdsIntegrateEX::writeRoutines(outfile);
  xmdsIntegrateARK45::writeRoutines(outfile);
}

// **************************************************************************
void xmdsIntegrateARK45EX::writeResetRoutine(
                                             FILE *const outfile) const {
  if (debugFlag) {
    printf("xmdsIntegrateARK45EX::writeResetRoutine\n");
  }

  const char *const fieldName = simulation()->field()->name()->c_str();

  const xmdsVector* mainVector;

  if (!simulation()->field()->getVector("main", mainVector)) {
    throw xmdsException("Internal error in xmdsIntegrateARK45::writeResetRoutine: cannot find 'main' vector");
  }

  const char* typeName="";
  if (mainVector->vectorType()==COMPLEX) {
    typeName="complex";
  }
  else if (mainVector->vectorType()==DOUBLE) {
    typeName="double";
  }

  fprintf(outfile, "/* **************************************************/\n");
  fprintf(outfile, "void _segment%li_reset(%s* _reset_to, double _step){\n", segmentNumber, typeName);
  if (simulation()->parameters()->useOpenMP) {
    // loop iterations are independent, so parallelise!
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp parallel for\n"
            "#endif\n");
  }
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "for (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", fieldName);
  }
  else {
    fprintf(outfile, "for (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", fieldName, fieldName);
  }
  fprintf(outfile, "    _%s_main[_i1] = _reset_to[_i1];\n", fieldName);
  fprintf(outfile, "}\n");
  fprintf(outfile, "\n");

}

// **************************************************************************
void xmdsIntegrateARK45EX::writeSingleStepCode(
                                               FILE *const outfile,
                                               const stepCaseEnum& stepCase) const {
  if (debugFlag) {
    printf("xmdsIntegrateARK45EX::writeSingleStepCode\n");
  }

  const char *const fieldName = simulation()->field()->name()->c_str();
  const char *const propDim = simulation()->parameters()->propDimName.c_str();

  list<XMLString> tempVectorNamesList;
  tempVectorNamesList.push_back("main");

  const char* indent = "  ";
  if (simulation()->parameters()->errorCheck) {
    indent = "    ";
  }

  const char* noiseVector="";
  if ((simulation()->parameters()->stochastic)&&(!noNoises())) {
    noiseVector=", _noise_vector";
  }

  simulation()->field()->vectors2space(outfile, 0, tempVectorNamesList, indent);

  if (simulation()->parameters()->useOpenMP) {
    // We begin a parallel region as we have two for loops following each other
    // This minimises the penalty of creating/destroying threads for silly implementations of OpenMP
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp parallel\n"
            "{\n"
            "#pragma omp for nowait\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// a_k=y2=y1\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _akfield_%s_main[_i1] = _%s_check[_i1] = _%s_main[_i1];\n", indent, fieldName, fieldName, fieldName);
  fprintf(outfile, "\n");

  if (simulation()->parameters()->useOpenMP) {
    // Continuing the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp for\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// a_i=y1\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _aifield_%s_main[_i1] = _%s_main[_i1];\n", indent, fieldName, fieldName);
  if (simulation()->parameters()->useOpenMP) {
    // End the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "}\n"
            "#endif\n");
  }
  fprintf(outfile, "\n");

  fprintf(outfile, "%s_active_%s_main = _akfield_%s_main;\n", indent, fieldName, fieldName);
  fprintf(outfile, "\n");

  if (usesKOperators()) {
    fprintf(outfile, "%s// calculate the co derivative terms\n", indent);
    fprintf(outfile, "%s_segment%li_calculate_co_terms();\n", indent, segmentNumber);
    fprintf(outfile, "\n");
  }

  fprintf(outfile, "%s// a_k=G[a_k, t]\n", indent);
  fprintf(outfile, "%s_segment%li_calculate_delta_a(_step, cycle%s);\n", indent, segmentNumber, noiseVector);
  fprintf(outfile, "\n");

  if (simulation()->parameters()->useOpenMP) {
    // We begin a parallel region as we have two for loops following each other
    // This minimises the penalty of creating/destroying threads for silly implementations of OpenMP
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp parallel\n"
            "{\n"
            "#pragma omp for nowait\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// y1=y1+c_1*a_k\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _%s_main[_i1] += _c[1]*_akfield_%s_main[_i1];\n", indent, fieldName, fieldName);

  if (simulation()->parameters()->useOpenMP) {
    // Continuing the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp for\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// y2=y2+cs_1*a_k\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _%s_check[_i1] += _cs[1]*_akfield_%s_main[_i1];\n", indent, fieldName, fieldName);

  if (simulation()->parameters()->useOpenMP) {
    // Continuing the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp for\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// a_k = a_i + b_21*a_k\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _akfield_%s_main[_i1] = _aifield_%s_main[_i1] + _b[2][1]*_akfield_%s_main[_i1];\n", indent, fieldName, fieldName, fieldName);
  if (simulation()->parameters()->useOpenMP) {
    // End the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "}\n"
            "#endif\n");
  }
  fprintf(outfile, "\n");

  fprintf(outfile, "%s%s += _a[2]*_step;\n", indent, propDim);
  fprintf(outfile, "\n");

  if (usesKOperators()) {
    fprintf(outfile, "%s// re-calculate the co derivative terms\n", indent);
    fprintf(outfile, "%s_segment%li_calculate_co_terms();\n", indent, segmentNumber);
    fprintf(outfile, "\n");
  }

  fprintf(outfile, "%s// a_k=G[a_k, t+aa_2*dt]\n", indent);
  fprintf(outfile, "%s_segment%li_calculate_delta_a(_step, cycle%s);\n", indent, segmentNumber, noiseVector);
  fprintf(outfile, "\n");

  if (simulation()->parameters()->useOpenMP) {
    // loop iterations are independent, so parallelise!
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp parallel for\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// c_2==cs_2==0\n", indent);

  fprintf(outfile, "%s// a_j = d_1*a_i + d_2*y1 + d_3*a_k \n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _ajfield_%s_main[_i1] = _d[1]*_aifield_%s_main[_i1] + _d[2]*_%s_main[_i1]+_d[3]*_akfield_%s_main[_i1] ;\n", indent, fieldName, fieldName, fieldName, fieldName);
  fprintf(outfile, "\n");

  fprintf(outfile, "%s%s += _a[3]*_step;\n", indent, propDim);
  fprintf(outfile, "\n");

  fprintf(outfile, "%s_active_%s_main = _ajfield_%s_main;\n", indent, fieldName, fieldName);
  fprintf(outfile, "\n");

  if (usesKOperators()) {
    fprintf(outfile, "%s// re-calculate the co derivative terms\n", indent);
    fprintf(outfile, "%s_segment%li_calculate_co_terms();\n", indent, segmentNumber);
    fprintf(outfile, "\n");
  }

  fprintf(outfile, "%s// a_j=G[a_j, t+aa_3*dt]\n", indent);
  fprintf(outfile, "%s_segment%li_calculate_delta_a(_step, cycle%s);\n", indent, segmentNumber, noiseVector);
  fprintf(outfile, "\n");

  if (simulation()->parameters()->useOpenMP) {
    // We begin a parallel region as we have two for loops following each other
    // This minimises the penalty of creating/destroying threads for silly implementations of OpenMP
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp parallel\n"
            "{\n"
            "#pragma omp for\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// a_l = e_1*a_i + e_2*y1 + e_3*a_k + e_4*a_j\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _alfield_%s_main[_i1] = _e[1]*_aifield_%s_main[_i1] + _e[2]*_%s_main[_i1] + _e[3]*_akfield_%s_main[_i1] + _e[4]*_ajfield_%s_main[_i1] ;\n", indent, fieldName, fieldName, fieldName, fieldName, fieldName);
  fprintf(outfile, "\n");

  if (simulation()->parameters()->useOpenMP) {
    // Continuing the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp for nowait\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// y1=y1+c_3*a_j\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _%s_main[_i1] += _c[3]*_ajfield_%s_main[_i1];\n", indent, fieldName, fieldName);

  if (simulation()->parameters()->useOpenMP) {
    // Continuing the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp for\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// y2=y2+cs_3*a_j\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _%s_check[_i1] += _cs[3]*_ajfield_%s_main[_i1];\n", indent, fieldName, fieldName);
  if (simulation()->parameters()->useOpenMP) {
    // End the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "}\n"
            "#endif\n\n");
  }

  fprintf(outfile, "%s%s += _a[4]*_step;\n", indent, propDim);
  fprintf(outfile, "\n");

  fprintf(outfile, "%s_active_%s_main = _alfield_%s_main;\n", indent, fieldName, fieldName);
  fprintf(outfile, "\n");

  if (usesKOperators()) {
    fprintf(outfile, "%s// re-calculate the co derivative terms\n", indent);
    fprintf(outfile, "%s_segment%li_calculate_co_terms();\n", indent, segmentNumber);
    fprintf(outfile, "\n");
  }

  fprintf(outfile, "%s// a_l=G[a_l, t+aa_4*dt]\n", indent);
  fprintf(outfile, "%s_segment%li_calculate_delta_a(_step, cycle%s);\n", indent, segmentNumber, noiseVector);
  fprintf(outfile, "\n");

  if (simulation()->parameters()->useOpenMP) {
    // We begin a parallel region as we have two for loops following each other
    // This minimises the penalty of creating/destroying threads for silly implementations of OpenMP
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp parallel\n"
            "{\n"
            "#pragma omp for nowait\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// y1=y1+c_4*a_j\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _%s_main[_i1] += _c[4]*_alfield_%s_main[_i1];\n", indent, fieldName, fieldName);

  if (simulation()->parameters()->useOpenMP) {
    // Continuing the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp for\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// y2=y2+cs_4*a_j\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _%s_check[_i1] += _cs[4]*_alfield_%s_main[_i1];\n", indent, fieldName, fieldName);

  if (simulation()->parameters()->useOpenMP) {
    // Continuing the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp for\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// a_l = f_1*a_i + f_2*y1 + f_3*a_k + f_4*a_j +f_5*a_l;\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _alfield_%s_main[_i1] = _f[1]*_aifield_%s_main[_i1] + _f[2]*_%s_main[_i1] + _f[3]*_akfield_%s_main[_i1] + _f[4]*_ajfield_%s_main[_i1] + _f[5]*_alfield_%s_main[_i1] ;\n", indent, fieldName, fieldName, fieldName, fieldName, fieldName, fieldName);
  if (simulation()->parameters()->useOpenMP) {
    // End the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "}\n"
            "#endif\n\n");
  }
  fprintf(outfile, "\n");

  fprintf(outfile, "%s%s += _a[5]*_step;\n", indent, propDim);
  fprintf(outfile, "\n");

  if (usesKOperators()) {
    fprintf(outfile, "%s// re-calculate the co derivative terms\n", indent);
    fprintf(outfile, "%s_segment%li_calculate_co_terms();\n", indent, segmentNumber);
    fprintf(outfile, "\n");
  }

  fprintf(outfile, "%s// a_l=G[a_l, t+aa_5*dt]\n", indent);
  fprintf(outfile, "%s_segment%li_calculate_delta_a(_step, cycle%s);\n", indent, segmentNumber, noiseVector);
  fprintf(outfile, "\n");


  if (simulation()->parameters()->useOpenMP) {
    // We begin a parallel region as we have two for loops following each other
    // This minimises the penalty of creating/destroying threads for silly implementations of OpenMP
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp parallel\n"
            "{\n"
            "#pragma omp for\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// c_5==0\n", indent);
  fprintf(outfile, "%s// y2=y2+cs_5*a_l\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _%s_check[_i1] += _cs[5]*_alfield_%s_main[_i1];\n", indent, fieldName, fieldName);

  if (simulation()->parameters()->useOpenMP) {
    // Continuing the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp for\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// a_l= g_1*a_i + g_2*a_k + g_3*a_j + g_4*y_1 +g_5*a_l +g_6*y_2;\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _alfield_%s_main[_i1] = _g[1]*_aifield_%s_main[_i1] + _g[2]*_akfield_%s_main[_i1] + _g[3]*_ajfield_%s_main[_i1] + _g[4]*_%s_main[_i1] + _g[5]*_alfield_%s_main[_i1] + _g[6]*_%s_check[_i1] ;\n", indent, fieldName, fieldName, fieldName, fieldName, fieldName, fieldName, fieldName);
  if (simulation()->parameters()->useOpenMP) {
    // End the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "}\n"
            "#endif\n\n");
  }
  fprintf(outfile, "\n");

  fprintf(outfile, "%s%s += _a[6]*_step;\n", indent, propDim);
  fprintf(outfile, "\n");

  if (usesKOperators()) {
    fprintf(outfile, "%s// re-calculate the co derivative terms\n", indent);
    fprintf(outfile, "%s_segment%li_calculate_co_terms();\n", indent, segmentNumber);
    fprintf(outfile, "\n");
  }

  fprintf(outfile, "%s// a_l=G[a_l, t+aa_6*dt]\n", indent);
  fprintf(outfile, "%s_segment%li_calculate_delta_a(_step, cycle%s);\n", indent, segmentNumber, noiseVector);
  fprintf(outfile, "\n");

  if (simulation()->parameters()->useOpenMP) {
    // We begin a parallel region as we have two for loops following each other
    // This minimises the penalty of creating/destroying threads for silly implementations of OpenMP
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp parallel\n"
            "{\n"
            "#pragma omp for nowait\n"
            "#endif\n");
  }
  fprintf(outfile, "%s// y1=y1+c_6*a_l\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _%s_main[_i1] += _c[6]*_alfield_%s_main[_i1];\n", indent, fieldName, fieldName);
  if (simulation()->parameters()->useOpenMP) {
    // Continuing the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "#pragma omp for\n"
            "#endif\n");
  }

  fprintf(outfile, "%s// y2=y2+cs_6*a_l\n", indent);
  if (simulation()->parameters()->usempi&!simulation()->parameters()->stochastic) {
    fprintf(outfile, "%sfor (long _i1=0; _i1<total_local_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName);
  }
  else {
    fprintf(outfile, "%sfor (long _i1=0; _i1<_%s_size*_%s_main_ncomponents; _i1++)\n", indent, fieldName, fieldName);
  }
  fprintf(outfile, "%s _%s_check[_i1] += _cs[6]*_alfield_%s_main[_i1];\n", indent, fieldName, fieldName);
  if (simulation()->parameters()->useOpenMP) {
    // End the parallel region
    fprintf(outfile, "#ifdef _OPENMP\n"
            "}\n"
            "#endif\n\n");
  }

  fprintf(outfile, "%s// t->t+dt\n", indent);
  fprintf(outfile, "%s%s -= _a[6]*_step;\n", indent, propDim);
  fprintf(outfile, "\n");

  fprintf(outfile, "%s_active_%s_main = _%s_main;\n", indent, fieldName, fieldName);
  fprintf(outfile, "\n");
}

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
